module F = PLTLFormula 
module G = PLTLGraph

let ischarnum s =
  if List.mem s ['0';'1';'2';'3';'4';'5';'6';'7';'8';'9'] then true else false;;

let getstatenum filename = 
  let fp = open_in filename in
  let rec read acc = 
    if ischarnum (input_line fp).[0] then
      let acc = acc + 1 in read acc
    else acc 
  in 
  let ret = read 0 in
  let _ = close_in fp in
    ret;;

let statecontent str =
  let rec isleftbracket str n = 
    if str.[n] = '{' then n else
      if n = (String.length str) - 1 then -1
      else isleftbracket str (n+1)
  in 
  let start = isleftbracket str 0 in
    if start = -1 then failwith "this line doesn't contain '{'"
    else String.sub str start ((String.length str) - start -1);;

let rec getlist al aw s n =
  if n > (String.length s) -1 then al
  else if s.[n] = ',' or s.[n] = ']' then 
    let al = al@[aw] in
    let aw = "" in
      getlist al aw s (n+1)
  else  
    if n = (String.length s) - 1 then al
    else 
      let aw = aw^(String.make 1 s.[n]) in
	getlist al aw s (n+1);;

let statelist str = getlist [] "" str 1;;

let getstateid str = 
  let rec isspace str n =
    if str.[n] = ' ' then n else
      if n = (String.length str) - 1 then -1
      else isspace str (n+1)
  in 
  let space = isspace str 0 in
    if space = -1 then failwith "this line doesn't contain a space"
    else int_of_string (String.sub str 0 space);;

let readtraces filename = 
  let fp = open_in filename in 
  let _ = while ischarnum (input_line fp).[0] do () done in
  let rec traces acc =
    try 
      let thisline = input_line fp in
      let acc = acc@[String.sub thisline 0 (String.length thisline -1)] in traces acc 
    with _ -> acc
  in 
  let ret = traces [] in
  let _ = close_in fp in
    ret;;

let rec tracelist acc = function
  | h::t -> 
      let thislist = getlist [] "" h 1 in 
      let acc = acc@[thislist] in
	tracelist acc t
  | [] -> acc;;

let mkstatehash filename = 
  let num = getstatenum filename in
  let hashtbl = Hashtbl.create num in
  let fp = open_in filename in
  let rec addhash ht n = 
    if n >= num then ht 
    else
      begin 
	let line = input_line fp in
	let content = statecontent line in
	let state = statelist content in
	let stateindex = getstateid line in
	let _ = Hashtbl.add ht stateindex state in
	  addhash ht (n+1)
      end 
  in 
  let ret = addhash hashtbl 0 in
  let _ = close_in fp in
    ret;;

let gettraces filename =
  let traces = readtraces filename in
  let tracel = tracelist [] traces in
  let rec tointll ll acc = 
    match ll with
	h::t -> let acc = acc@[(List.map (fun x -> int_of_string x) h)] in tointll t acc
      | [] -> acc
  in 
    tointll tracel [];;

let traces = gettraces "traces.txt";;

let statehash = mkstatehash "traces.txt";;

(*if e is the nth member of l, which is a list of int, return n*)
let listnth l e =
  let rec findmem acc l e =
    match l with
	h::t -> if h = e then acc else findmem (acc+1) t e
      | [] -> -1
  in 
    findmem 0 l e;;

(*return the next state id on trace, return -1 if no next state.
since the last state will always be repeating, it is ignored   *)
let nextstateid stateid trace =
  let nth = listnth trace stateid in
    if (nth + 1) >= ((List.length trace) - 1) then -1 else List.nth trace (nth + 1);;

(*return the first state id on trace, return -1 if trace is empty*)
let fststateid trace =
  match trace with
      h::t -> h
    | [] -> -1;;

(*if atomic proposition is in this state, return true, otherwise 
return false. stateid is int                                   *)
let eval ap stateid = 
  let aps = F.dest_ap ap in
  let state = Hashtbl.find statehash stateid in
    if List.mem aps state then true else false;;

(*test if a subset propositional formula is satisfiable at stateid 
on trace, where stateid is int and trace is a list of int       *)
let rec testsubf f stateid trace =
  (*let _ = print_string ("testing "^(F.exportFormula f)^" at state ");print_int stateid; print_endline "" in*)
  match f with 
      F.TRUE -> true   
    | F.FALSE -> false
    | F.AP _ ->  eval f stateid 
    | F.NOT f1 -> not (testsubf f1 stateid trace)
    | F.EQU (f1, f2) -> (not (testsubf f1 stateid trace) or (testsubf f2 stateid trace)) & (not (testsubf f2 stateid trace) or (testsubf f1 stateid trace)) 
    | F.IMP (f1, f2) -> not (testsubf f1 stateid trace) or (testsubf f2 stateid trace)
    | F.OR (f1, f2) -> (testsubf f1 stateid trace) or (testsubf f2 stateid trace)
    | F.AND (f1, f2) -> (testsubf f1 stateid trace) & (testsubf f2 stateid trace) 
    | F.X f1 -> (*if f1 is true in next state then return true, else false*)
	let nextid = nextstateid stateid trace in
	  if nextid = -1 then false else testsubf f1 nextid trace 
    | F.EV f1 -> 
	let rec testev f1 stateid trace =
	  if testsubf f1 stateid trace then true (*if f1 is true at this state, return true*)
	  else (*if all states are tested, then return false else test the next state*)
	    let nextid = nextstateid stateid trace in  
	      if nextid = -1 then false else testev f1 nextid trace  
	in 
	  testev f1 stateid trace
    | F.AW f1 ->
	let rec testaw f1 stateid trace =
	  if testsubf f1 stateid trace then (*if f1 is true at this state, then test the next state*)
	    let nextid = nextstateid stateid trace in (*if all states are tested then return true*) 
	      if nextid = -1 then true else testaw f1 nextid trace 
	  else (*if f1 is false at this state, return false*)
	    false
	in  
	  testaw f1 stateid trace
    | F.UN (f1, f2) ->
	let rec testun f1 f2 beginid nowid trace =
	  if testsubf f2 nowid trace then (*if f2 is true at current state*)
	    let rec testf1 f1 fromid nowid trace = (*then test if f1 is always true from begin to current*)
	      if testsubf f1 fromid trace then  (*if f1 is true at this state*)
		let nextid = nextstateid fromid trace in 
		  if nextid  = nowid then true  (*if the next state is current state then return true*)
		  else testf1 f1 nextid nowid trace (*else test the next state*)
	      else false (*if f1 is not true at this state, return false*)
	    in (*as long as f1 until f2 is true, return true*) 
	      if testf1 f1 beginid nowid trace then true  
	      else (*if f1 is not always true from begin to now*)
		let nextid = nextstateid nowid trace in (*try next state*)
		  if nextid = -1 then false else 
		    testun f1 f2 beginid nextid trace 
	  else (*if f2 is not true at current state, try next state*)
	    let nextid = nextstateid nowid trace in
	      if nextid = -1 then false else (*if all states tried, return false*)
		testun f1 f2 beginid nextid trace
	in  
	  testun f1 f2 stateid stateid trace
    | F.BF (f1, f2) ->
	let rec testbf f1 f2 beginid nowid trace =
	  if testsubf f2 nowid trace then (*if f2 is true at current state*)
	    let rec testf1 f1 fromid nowid trace = (*then test if f1 is true somewhere before*)
	      if testsubf f1 fromid trace then true  (*if f1 is true at this state then return true*)
	      else (*else try if f1 is true at next state*)
		let nextid = nextstateid fromid trace in 
		  if nextid = nowid then false  (*if next state is current state, return false*)
		  else testf1 f1 nextid nowid trace 
	    in (*if f1 is true somewhere before current state, then test next state*)
	      if testf1 f1 beginid nowid trace then 
		let nextid = nextstateid nowid trace in
		  testbf f1 f2 beginid nextid trace
	      else false (*if f2 is true at current state but f1 is always false before that, return false*)
	  else  (*if f2 is false at current state, try next state*)
	    let nextid = nextstateid nowid trace in
	      if nextid = -1 then true else (*if all states tested, and f2 is always false, return true*)
		testbf f1 f2 beginid nextid trace 
	in  
	  testbf f1 f2 stateid stateid trace
    | _ -> failwith "undefined LTL operator";;

	
(*test if f is satisfiable in any of the traces. statehash is a hashtable 
with int*(string list), where int for stateid and string list for the content
 of the state. traces is a list of list of int, contains all traces*)
let isSat f =
  match f with
      F.TRUE -> true
    | F.FALSE -> false
    | _ ->
	let rec testf f traces = 
	  match traces with
	      h::t -> 
		let fstid = fststateid h in
		  if fstid <> -1 then 
		    if testsubf f fstid h then true else 
		      testf f t
		  else 
		    testf f t
	    | [] -> false
	in 
	  testf f traces;;
