module P = PLTLFormula
module G = GenAndComp

let rec list_bigOp op base f = function
  | h::t -> op (f h) (list_bigOp op base f t)
  | [] -> base


(* actions, goals, results are strings
   resultexclusion is list of strings
   initial is formula
   actionmap is (string, formula list), formula should only be (negated) literals
*)
let mkBusProc actionsin goals results initial actionmap resultexclusion =
  let p s = P.AP s in
  let actions = "end"::("start"::actionsin) in
  let interleave a =
    let rec negs = function
      | h::t -> 
	  if h = a then negs t
	  else P.AND (P.NOT (p h),negs t)
      | [] -> P.TRUE
    in
      P.IMP(p a, negs actions)
  in
  let interleavef = list_bigOp P.const_and P.TRUE interleave actions in
  let progressf = list_bigOp P.const_or P.FALSE p actions in
  let mutex l = 
    let m2 a b = if a = b then P.TRUE else P.NOT (P.AND (p a, p b)) in
    let m1 a = list_bigOp P.const_and P.TRUE (m2 a) l in
      list_bigOp P.const_and P.TRUE m1 l
  in
  let mutexf = list_bigOp P.const_and P.TRUE mutex resultexclusion in
  let causality1 (a,s) = P.IMP(p a, (list_bigOp P.const_or P.FALSE (fun x->x) s )) in	
  let causality1f = list_bigOp P.const_and P.TRUE causality1 actionmap in
  let causality2 r =
    let eqigneg r f =
      match f with
	| P.AP s
	| P.NOT (P.AP s) -> r = s
	| _ -> false
    in
    let rec relevantactions base = function
      | (ha,hS)::t ->
	  if (List.exists (fun x -> x = ha) base) then
	    relevantactions base t
	  else if List.exists (eqigneg r) hS then
	    relevantactions (ha::base) t
	  else
	    relevantactions base t
      | [] -> base
    in
    let xs = (list_bigOp P.const_or P.FALSE p (relevantactions [] actionmap)) in
	P.AND(P.IMP (p r, (P.BF(P.X xs, P.NOT (p r)))),
	    P.IMP (P.NOT (p r), (P.BF (P.X xs, p r))))
  in
  let causality2f = list_bigOp P.const_and P.TRUE causality2 results in
  let once a = if a = "end" then P.TRUE else P.IMP (p a, P.X (P.AW (P.NOT (p a)))) in
  let oncef = list_bigOp P.const_and P.TRUE once actions in
  let finalf = P.IMP(p "end", P.AW (p "end")) in
  let goalf = list_bigOp P.const_or P.FALSE p goals in
  let initialf = P.IMP(p "start", list_bigOp P.const_and P.TRUE (fun x->x) initial) in
  let fs = [p "start"; P.AW initialf; P.EV goalf; P.EV (p "end"); P.AW interleavef; P.AW progressf; P.AW mutexf; P.AW causality1f; P.AW causality2f; P.AW oncef; P.AW finalf] in
    (list_bigOp P.const_and P.TRUE (fun x->x) fs)


let readDomKlg filename = 
  let fp = open_in filename in
  let astr = input_line fp in
  let rstr = input_line fp in
  let mstr = input_line fp in
  let estr = input_line fp in
  let gstr = input_line fp in 
  let istr = input_line fp in
    close_in fp;
    let p s = P.AP s in
    let np s = P.NOT (p s) in
    let actions = 
      let a = String.sub astr 2 ((String.length astr)-2) in
	Str.split (Str.regexp ";") a in
    let results = 
      let r = String.sub rstr 2 ((String.length rstr)-2) in
	Str.split (Str.regexp ";") r in
    let actionmaps = 
      let m = String.sub mstr 2 ((String.length mstr)-2) in
      let ml = Str.split (Str.regexp ";") m in
	List.map (fun w -> 
		    let act = Str.split (Str.regexp ",") w in  
		    let roa = List.map (fun z -> if (String.contains z '~') then (np (String.sub z 1 ((String.length z)-1))) else (p z)) (List.tl act) in 
		      (List.hd act, roa)) ml in
    let execlusion = 
      let e = String.sub estr 2 ((String.length estr)-2) in
      let el = Str.split (Str.regexp ";") e in
	List.map (fun x -> List.tl (Str.split (Str.regexp ",") x)) el in
    let goal = 
      let g = String.sub gstr 2 ((String.length gstr)-2) in
	Str.split (Str.regexp ";") g in
    let initial = 
      let i = String.sub istr 2 ((String.length istr)-2) in
	List.map (fun y -> if (String.contains y '~') then (np (String.sub y 1 ((String.length y)-1))) else (p y)) (Str.split (Str.regexp ";") i) in
      mkBusProc actions goal results initial actionmaps execlusion

let printDomKlg  =
  let fp = open_in "constraints.pltl" in
  let constraints = input_line fp in
  let _ = close_in fp in
  let dkstr = P.exportFormula (readDomKlg "DomainKnowledge.txt") in
  let fp = open_out "encoding.pltl" in
  let _ = output_string fp (dkstr^" & "^constraints);close_out fp in
  let fp = open_out "dk.pltl" in
    output_string fp dkstr;close_out fp 

