module F = PLTLFormula
module G = PLTLGraph
module T = PLTLTree
module B = Busproc

let rec printflist = function
  |[] -> print_string ""
  |h::t -> print_string ((F.exportFormula h)^";"); printflist t;; 

let rec printflistlist = function
  |[] -> printflist [] 
  |h::t -> print_string "["; printflist h; print_string"]"; printflistlist t;; 

(*read rules from file, and convert into a list of formula*)
let readrules filename = 
  let fp = open_in filename in
  let constraints = input_line fp in
  let _ = close_in fp in
  let formulae = F.importFormula constraints in
  let rec parserules f acc = 
    match f with
	F.AND (f1, f2) -> (parserules f1 acc)@(parserules f2 acc)
      | _ -> let acc = f::acc in acc
  in 
    parserules formulae [];;

(*convert a list of formulae into a big conjunction*)
let rec and_of_list fl = 
  match fl with 
      h::t -> F.AND (h, and_of_list t)
    | _ -> F.TRUE;;

(*find MUS by deletion of a set of formulae*)	  
let findmus f =
  let rec testsat head tail = 
    match tail with 
	h::t ->
	  if G.isSat (and_of_list (head@t)) then 
	    let _ = print_endline ((F.exportFormula h)^" is the root cause") in 
	      testsat (head@[h]) t 
	  else 
	    let _ = print_endline ((F.exportFormula h)^" is not the root cause") in 
	      testsat head t
      | _ -> head
  in
    testsat [] f;;

(*find MUS by deletion between domain knowledge and rules*)	  
let findmusdk dk rules =
  let rec testsat head tail = 
    match tail with 
	h::t ->
	  if G.isSat (F.AND ((and_of_list (head@t)), (and_of_list dk))) then 
	    let _ = print_endline ((F.exportFormula h)^" is the root cause") in 
	      testsat (head@[h]) t 
	  else 
	    let _ = print_endline ((F.exportFormula h)^" is not the root cause") in 
	      testsat head t
      | _ -> head
  in
    testsat [] rules;;
(*find the powerset of a list*)
let rec listpowerset = function
  | [] ->[[]]
  | h::t -> List.fold_left (fun xs t -> (h::t)::t::xs) [] (listpowerset t);;

(*find minimum UC of a set of formulae*)
let findmmus f =
  let powerf = listpowerset f in
  let rec testsat s n = 
    match s with
	h::t -> 
	  if List.length h = n then 
	    let _ = print_string "testing subset " in let _ = printflist h in let _ = print_endline "" in
	      if G.isSat (and_of_list h) then testsat t n else h
	  else testsat t n 
      | [] -> []
  in 
  let rec findmm s acc =
    let r = testsat s acc in
      match r with
	  [] -> if acc < List.length f then findmm s (acc+1) else [] 
	| _ -> r
  in 
    findmm powerf 1
 
(*find minimum UC of a set of rules with domain knowledge*)
let findmmusdk dk f =
  let powerf = listpowerset f in
  let rec testsat s n = 
    match s with
	h::t -> 
	  if List.length h = n then 
	    let _ = print_string "testing subset with DK " in let _ = printflist h in let _ = print_endline "" in
	      if G.isSat (F.AND ((and_of_list h), (and_of_list dk))) then testsat t n else h
	  else testsat t n 
      | [] -> []
  in 
  let rec findmm s acc =
    let r = testsat s acc in
      match r with
	  [] -> if acc < List.length f then findmm s (acc+1) else [] 
	| _ -> r
  in 
    findmm powerf 1

let rec flists_intersect acc l1 l2 =
  match l1 with
      h::t -> if List.mem h l2 then let acc = h::acc in flists_intersect acc t l2 else flists_intersect acc t l2
    | _ -> acc;; 

let inconanalysis stage opt =
  let dk = readrules "dk.pltl" in
  let rules = readrules "constraints.pltl" in
  let gamma = readrules "encoding.pltl" in   
  let start = Unix.gettimeofday () in 
    (*let mus = findmusdk unsatrules in  
      let _ = print_string "unsatdk is: " in let _ = printflist unsatdk in let _ = print_endline "" in
      let _ = print_string "unsatrule is: " in let _ = printflist unsatrules in let _ = print_endline "" in*)
  let analysis stage opt dk rules gamma =
    match opt with
      |1 -> 
	 begin
	   match stage with
	     |1 -> 
		let sat = (PLTLTree.isSat ~verbose:true (and_of_list dk)) in 
		  if sat then let _ = print_endline "Domain knowledge is satisfiable." in exit(0)
		  else 
		    let unsatcore = readrules "unsatcore.pltl" in
		      findmus unsatcore
	     |2 -> 
		let sat = (PLTLTree.isSat ~verbose:true (and_of_list rules)) in 
		  if sat then let _ = print_endline "Rules are satisfiable." in exit(0)
		  else 
		    let unsatcore = readrules "unsatcore.pltl" in 
		      findmus unsatcore
	     |3 -> 
		let sat = (PLTLTree.isSat ~verbose:true (and_of_list gamma)) in 
		  if sat then let _ = print_endline "Domain knowledge and rules are satisfiable." in exit(0)
		  else 
		    let unsatcore = readrules "unsatcore.pltl" in 
		      findmus unsatcore
	     |_ -> exit(1)
	 end
      |2 ->
	 begin
	   match stage with
	     |1 -> 
		let sat = (PLTLTree.isSat ~verbose:true (and_of_list dk)) in 
		  if sat then let _ = print_endline "Domain knowledge is satisfiable." in exit(0)
		  else 
		    let unsatcore = readrules "unsatcore.pltl" in
		      findmmus unsatcore
	     |2 -> 
		let sat = (PLTLTree.isSat ~verbose:true (and_of_list rules)) in 
		  if sat then let _ = print_endline "Rules are satisfiable." in exit(0)
		  else 
		    let unsatcore = readrules "unsatcore.pltl" in 
		      findmmus unsatcore
	     |3 -> 
		let sat = (PLTLTree.isSat ~verbose:true (and_of_list gamma)) in 
		  if sat then let _ = print_endline "Domain knowledge and rules are satisfiable." in exit(0)
		  else 
		    let unsatcore = readrules "unsatcore.pltl" in 
		      findmmus unsatcore
	     |_ -> exit(1)
	 end
      |3 -> 
	 begin
	   match stage with
	     |1 -> 
		let sat = (PLTLTree.isSat ~verbose:true (and_of_list dk)) in 
		  if sat then let _ = print_endline "Domain knowledge is satisfiable." in exit(0)
		  else findmmus dk
	     |2 -> 
		let sat = (PLTLTree.isSat ~verbose:true (and_of_list rules)) in 
		  if sat then let _ = print_endline "Rules are satisfiable." in exit(0)
		  else findmmus rules
	     |3 -> 
		let sat = (PLTLTree.isSat ~verbose:true (and_of_list gamma)) in 
		  if sat then let _ = print_endline "Domain knowledge and rules are satisfiable." in exit(0)
		  else findmmus gamma
	     |_ -> exit(1)
	 end
      |4 -> 
	 begin 
	   match stage with 
	     |3 -> 
		let sat = (PLTLTree.isSat ~verbose:true (and_of_list gamma)) in 
		  if sat then let _ = print_endline "Domain knowledge and rules are satisfiable." in exit(0)
		  else
 		    let unsatcore = readrules "unsatcore.pltl" in 
		    let unsatrules = flists_intersect [] unsatcore rules in
		    let unsatdk = flists_intersect [] unsatcore dk in
		      findmusdk unsatdk unsatrules
	     |_ -> exit(1)
	 end 
      |5 -> 
	 begin 
	   match stage with 
	     |3 -> 
		let sat = (PLTLTree.isSat ~verbose:true (and_of_list gamma)) in 
		  if sat then let _ = print_endline "Domain knowledge and rules are satisfiable." in exit(0)
		  else
 		    let unsatcore = readrules "unsatcore.pltl" in 
		    let unsatrules = flists_intersect [] unsatcore rules in
		    let unsatdk = flists_intersect [] unsatcore dk in
		      findmmusdk unsatdk unsatrules
	     |_ -> exit(1)
	 end 
      |6 -> 
	 begin 
	   match stage with 
	     |3 -> 
		let sat = (PLTLTree.isSat ~verbose:true (and_of_list gamma)) in 
		  if sat then let _ = print_endline "Domain knowledge and rules are satisfiable." in exit(0)
		  else findmmusdk dk rules
	     |_ -> exit(1)
	 end 
      |_ -> exit(1)
  in 
  let mus = analysis stage opt dk rules gamma in   
  let finish = Unix.gettimeofday () in  
    if List.length mus > 0 then 
      begin
	print_string "Root cause of inconsistency is: \n"; 
	printflist mus; 
	print_string "\n";
	print_endline ("Time used: "^(string_of_float (finish -. start))^" seconds")
      end 
    else 
      begin 
	print_endline "there is no MUS";
	print_endline ("Time used: "^(string_of_float (finish -. start))^" seconds")
      end     
