(** Accepts formulae from the standard input and tests them for satisfiability.
    Each formula has to be on a separate line and empty lines are ignored.
    The input is terminated by EOF.
    @author Florian Widmann
 *)


let _ = 
  let choice =
	if Array.length Sys.argv < 2 then 
	  let _ = print_endline "Please specify 'gr' or 'tr'" in exit(1)
	else Sys.argv.(1)
  in
  let _ = 
	match choice with
	  | "gr" | "tr" | "bj" -> ()
	  | _ -> let _ = print_endline "Please specify 'gr' or 'tr'" in exit(1)
  in
  let verb = (Array.length Sys.argv) - 2 in
  let prtgrph = if verb >= 2 then Some Sys.argv.(3)
	else None
  in
    let printRes sat =
      if verb = 0 then
        print_endline (if sat then "satisfiable\n" else "unsatisfiable\n")
      else ()
    in
    try
      while true do
        let input = read_line () in
        if not (GenAndComp.isEmptyString input) then
          let f = PLTLFormula.importFormula input in
		  if PLTLGraph.isSat ~verbose:verb ~print:prtgrph f then
			exit 1
		  else
			exit 0
        else ()
      done
    with End_of_file -> ()
