(** This module generates CTL formulae and runs solvers on them.
    The generated formulae contain randomly generated formulae
    as well as instances of certain classes of formulae.
    @author Florian Widmann
 *)


module G = GenAndComp
module P = PLTLFormula



let lFF = [(P.const_not, 1); (P.const_x, 1);
           (P.const_ev, 1); (P.const_aw, 1);]

let lFFF = [(P.const_or, 1); (P.const_and, 1);
            (P.const_un, 1); (P.const_bf, 1);]

let genFormula litFkt size =
  let nrLits = litFkt size in
  let lF =
    G.genLiterals (* ~litlst:[P.const_true; P.const_false] ~prio:10 *) P.const_ap "p" nrLits
  in
  let genRF = fst (G.mk_generator lF lFF lFFF [] [] [] [] []) in
  genRF size


let solvers =
(*  let dummy _ _ = (Some (false, 0.), 0) in*)
  let solveTreeBackjump f t =
	try
      let ret = G.evaluateFkt PLTLTree.isSat f t in
		(match ret with
		   |	(None,_) -> print_endline("Timeout! " ^ (P.exportFormula f))
		   |	_ -> ());
		ret
	with _ ->
	  begin
		print_endline ("Failed for " ^ (P.exportFormula f));
		(None,0)
	  end
  in
  let solveTreeGraph f t =
	G.evaluateFkt PLTLGraph.isSat f t
  in
  let solveTreeSchwendimann f =
	G.evaluateFkt PLTLTree_.isSat f
  in
  [(solveTreeBackjump, "Backjumping");(solveTreeGraph, "Graph");(solveTreeSchwendimann, "Schwendimann")]


(*
let seed =
  Random.self_init (); 
  Random.int ((1 lsl 30) - 1)

(* let seed = 308592204 *)

let _ = 
  Random.init seed;
  print_endline (string_of_int (seed))


let rate = int_of_string Sys.argv.(1)
let nr = int_of_string Sys.argv.(2)
let start = int_of_string Sys.argv.(3)
let stop = int_of_string Sys.argv.(4)
let step = int_of_string Sys.argv.(5)

let litFkt n = n / rate 

let _ = G.enumSolver ~dots:10 ~timeout:100 P.exportFormula (genFormula litFkt) solvers nr start stop step
*)


let logicFkts = {
  G.l_tt = P.TRUE;
  G.l_ff = P.FALSE;
  G.l_ap = (fun s i -> P.AP (s ^ (string_of_int i)));
  G.l_not = P.const_not;
  G.l_and = P.const_and;
  G.l_or = P.const_or;
  G.l_imp = P.const_imp;
  G.l_aa = (fun s i -> if s = "a" && i = 0 then () else raise (P.PLTLException "not implemented"));
  G.l_box = (fun _ f -> P.X f);
  G.l_dia = (fun _ f -> P.X f);
  G.l_aw = (fun _ f -> P.AW f);
  G.l_ev = (fun _ f -> P.EV f);
  G.l_until = P.const_un
}

let fs =
  let file = open_in "/tmp/benchmarks/listing.txt" in
  let rec readf acc =
	try
	  readf ((input_line file)::acc)
	with End_of_file -> acc
  in
  let ret = readf [] in
	close_in file;
	ret

let getF fs n =
  let filename = List.nth fs n in
	print_endline filename;
  let file = open_in filename in
  let f = P.importFormula_ic file in
  close_in file;
  f

let _ = G.enumSolver ~dots:10 ~timeout:60 P.exportFormula (getF fs) solvers 1 0 (List.length fs) 1
