module C = Cartesian
module G = PLTLGraph 
module F = PLTLFormula 
module T = PLTLTrace

let rec printlist = function
  |[] -> print_string ""
  |h::t -> print_string (h^";"); printlist t 

let rec printlistlist = function
  |[] -> printlist [] 
  |h::t -> print_string "["; printlist h; print_string"]"; printlistlist t 

let rec sublist list i j = 
  if i > j then 
    []
  else 
    if i = j then
      [List.nth list i]
    else 
      (List.nth list i)::(sublist list (i+1) j)   

let rec outputlist file = function
  |h::t -> output_string file (h^";"); outputlist file t
  |[] -> ()
 
let rec outputlistlist file = function 
  |h::t -> output_string file "["; outputlist file h; output_string file "];"; outputlistlist file t
  |[] -> () 

let rec overlaplist = function
  |head::tail -> 
     let rec testrest act = function
       |h::t ->
	  if (String.compare act h) == 0 then true
	  else testrest act t
       |[] -> false
     in
       if testrest head tail then true
       else overlaplist tail
  |[] -> false
     
let isemptylist = function
  |[] -> true
  |h::t -> false 

let procMine = 
  let starttime = Unix.gettimeofday () in
  let _ = print_endline "reading domain knowledge" in
  let fp = open_in "DomainKnowledge.txt" in 
  let astr = input_line fp in 
  let rstr = input_line fp in  
  let mstr = input_line fp in  
  let estr = input_line fp in
  let _ = input_line fp in
  let istr = input_line fp in
  let _ = close_in fp in 
  let p s = F.AP s in
  let np s = F.NOT (p s) in
  let actions = 
    let a = String.sub astr 2 ((String.length astr)-2) in
      Str.split (Str.regexp ";") a in
  let allactions = "start"::"end"::actions in
  let results = 
    let r = String.sub rstr 2 ((String.length rstr)-2) in
      Str.split (Str.regexp ";") r in
  let actionmaps = 
    let m = String.sub mstr 2 ((String.length mstr)-2) in
    let ml = Str.split (Str.regexp ";") m in
      List.map (fun w -> 
		  let act = Str.split (Str.regexp ",") w in  
		  let roa = List.map (fun z -> if (String.contains z '~') then (np (String.sub z 1 ((String.length z)-1))) else (p z)) (List.tl act) in 
		    (List.hd act, roa)) ml in
  let execlusionwa = 
    let e = String.sub estr 2 ((String.length estr)-2) in
      Str.split (Str.regexp ";") e in
  let execlusion = List.map (fun x -> List.tl (Str.split (Str.regexp ",") x)) execlusionwa in 
  let initials = 
    let i = String.sub istr 2 ((String.length istr)-2) in
    let ilist = Str.split (Str.regexp ";") i in
      List.filter (fun x -> if x.[0] = '~' then false else true) ilist in
  let _ = print_endline "testing optional actions" in
  let rec testOpt acc = function 
    |h::t ->
       let acc =
	 let _ = print_endline ("testing G~"^h) in
	   if (T.isSat (F.importFormula ("G~"^h))) then 
 	     h::acc 
 	   else  
 	     acc
       in
	 testOpt acc t 
    |[] -> acc 
  in 
  let optacts = testOpt [] actions in
  let _ = print_endline "optional actions are:" in
  let _ = if List.length optacts = 0 then print_endline "no optional actions" else printlist optacts;print_endline "" in
  let isOpt act = if (List.mem act optacts) then true else false in
  let rec getmandex acc = function 
    |h::t->
       let head = List.hd (Str.split (Str.regexp ",") h) in
       let tail = List.tl (Str.split (Str.regexp ",") h) in
       let acc =
	 if (isOpt head) then
	   acc
	 else 
	   tail::acc
       in
	 getmandex acc t
    |[] -> acc 
  in
  let execlusionm = getmandex [] execlusionwa in
  let _ = print_endline "generating all combinations of results" in
  let rim = C.cartesian execlusionm in
  let ri = C.cartesian execlusion in 
  let _ = print_endline "testing proper execution" in
  let rec isProper optact = function
    |head::tail -> 
       begin 	  
	 let rec testlist acc = function
	   |h::t ->
	      let acc = acc^h^"&" in 
		testlist acc t
	   |[] -> acc in
	 let thistest = testlist "" head in
	   begin 
	     let _ = print_endline ("testing F(" ^ thistest^optact^")") in
	       if (T.isSat (F.importFormula ("F("^thistest^optact^")"))) 
	       then isProper optact tail  
	       else true 
	   end  
       end
    |[] -> false 
  in
  let rec testProper = function
    |h::t ->
       begin 
	 if (isProper h ri) then true
	 else 
	   let _ = print_endline (h^" does not have proper execution") in 
	     exit(1)
       end; 
	testProper t
    |[] -> () 
  in 
  let _ = testProper optacts in
  let _ = print_endline "testing data completeness" in
  let rec isComplete = function
    |head::tail -> 
       begin 
	 let rec testlist acc = function
	   |h::t ->
	      let acc = acc^h^"&" in
		testlist acc t
	   |[] -> acc in
	 let thistest = testlist "" head in
	   begin 
	     let _ = print_endline ("testing F(" ^ thistest^"True)") in
	       if (T.isSat (F.importFormula ("F("^thistest^" True)")))
	       then isComplete tail 
	       else let _ = print_endline (thistest^" is not satisfiable, data is not complete") in
		 exit(1)
	   end
       end 
    |[] -> true 
  in
  let testComplete = isComplete rim in
  let _ = print_endline "testing order relations" in
  let unneg f = 
    match f with 
	F.NOT a -> a
      | _ -> f
  in 
  let rec findResults acc act = function
    |h::t ->
       if ((String.compare act (fst h)) == 0) && not (List.mem (F.exportFormula (unneg (List.hd (snd h)))) initials) then
	 let acc = acc@(snd h) in
	   findResults acc act t
       else
	 findResults acc act t
    |[] -> acc 
  in
  let rec testactr a b = function
    | h::t -> 
	if T.isSat (F.importFormula ("F("^a^" & X ( "^b^" & "^(F.exportFormula h)^"))")) then
	  testactr a b t
	else false
    | [] -> true
  in
  let rec testOrder rel = function
    |head::tail ->  
       begin
	 let rec thistest acc act = function 
	   |h::t ->
	      begin 
		let testRelation a b = 
		  let _ = print_endline ("testing relations between "^a ^" and "^b) in
		  let resultb = findResults [] b actionmaps in
		  let resulta = findResults [] a actionmaps in
		  let sat1 = 
		    if resultb = [] then
		      T.isSat (F.importFormula ("F("^a^" & X "^b^")")) 
		    else 
		      testactr a b resultb
		  in
		  let sat2 = 
		    if resulta = [] then
		      (T.isSat (F.importFormula ("F("^b^" & X "^a^")"))) 
		    else 
		      testactr b a resulta
		  in
		    if ((not sat1) && (not sat2)) then
		      [a;"#";b] 
		    else 
		      begin 
			if (sat1 && (not sat2)) then
			  [a;">>";b]
			else 
			  begin 
			    if ((not sat1) && sat2) then
			      [b;">>";a]
			    else 
			      [a;"||";b]
			  end
		      end in 
		let acc = (testRelation act h)::acc in 
		  thistest acc act t
	      end 
	   |[] -> acc in
	 let thisactrel = thistest [] head tail in
	 let rel = rel@thisactrel in
	   testOrder rel tail
       end 
    |[] -> rel 
  in
  let relations = testOrder [] allactions in
  let rec findgg acc c relations =
    match relations with
	h::t -> 
	  if (String.compare (List.nth h 1) ">>") == 0 && (String.compare (List.nth h 0) c) == 0 then
	    let acc = acc@[(List.nth h 2)] in
	      findgg acc c t
	  else 
	    findgg acc c t
      | [] -> acc
  in 
  let rec notgg c b =
    let crel = findgg [] c relations in
      match crel with
	  h::t ->
	    if (String.compare h b) == 0 then
	      false
	    else 
	      if notgg h b then
		let rec notggl cl b =
		  match cl with
		      h::t -> 
			if notgg h b then notggl t b
			else false 
		    | [] -> true
		in 
		  notggl t b
	      else 
		false
	| [] -> true
  in  
  let dirgg a b =
    let arel = findgg [] a relations in
    let rec testdir al b =
      match al with
	  h::t -> 
	    if notgg h b then testdir t b
	    else false
	| [] -> true
    in 
      testdir arel b
  in 
  let rec testDirOrder acc relations =
    match relations with
	h::t -> 
	  if (String.compare (List.nth h 1) ">>") == 0 then
	    if dirgg (List.nth h 0) (List.nth h 2) then
	      let acc = [(List.nth h 0);"->";(List.nth h 2)]::acc in
		testDirOrder acc t
	    else
	      let acc = h::acc in
		testDirOrder acc t
	  else 
	    let acc = h::acc in
	      testDirOrder acc t
      | [] -> acc
  in
  let relations = testDirOrder [] relations in
  let _ = print_endline "testing precedence conditions" in
  let rec testCond rcond = function
    |head::tail -> 
       if ((String.compare (List.nth head 1) "->") == 0) then
	 begin
	   let resultf = findResults [] (List.nth head 0) actionmaps in
	   let rec findCond acc rel = function
	     |h::t ->
		let _ = print_endline ("testing F("^(List.nth rel 0)^" & "^(F.exportFormula h)^" & X "^(List.nth rel 2)^")") in
		  if (T.isSat (F.importFormula ("F("^(List.nth rel 0)^" & "^(F.exportFormula h)^" & X "^(List.nth rel 2)^")"))) then
		    let acc = acc@[(F.exportFormula h)] in
		      findCond acc rel t
		  else
		    findCond acc rel t
	     |[] -> acc 
	   in
	     if (resultf != []) then
	       let relcond = findCond head head resultf in
	       let rcond = relcond::rcond in
		 testCond rcond tail
	     else 
	       let rcond = head::rcond in
		 testCond rcond tail
	 end
       else
	 let rcond = head::rcond in
	   testCond rcond tail
    |[] -> rcond 
  in
  let relationscond = testCond [] relations in   
  let writeresults =  
    let _ = print_endline "writing analysis results" in
    let fp = open_out_gen [Open_wronly; Open_append; Open_creat] 0o600 "output.bpm" in
      output_string fp "Optional actions are:\n";
      outputlist fp optacts;
      output_string fp "\nRelations between actions are:\n";
      outputlistlist fp relationscond;
      output_string fp "\n\n";
      close_out fp
  in
  let writegraphscript = 
    let _ = print_endline "writing graph script" in
    let fp = open_out "graphscript.dot" in
    let _ = output_string fp "digraph G{\n rankdir = LR\n" in
    let _ = output_string fp "start [label = \"start\", shape = circle]\n" in
    let _ = output_string fp "end [label = \"end\", shape = circle, style = bold]\n" in
    let rec writeactions = function
      |h::t ->
	 output_string fp (h^" [label = \""^h^"\", shape = box, style = rounded]\n");
	  writeactions t
      |[] -> () in
    let _ = writeactions actions in
    let rec writerelations = function
      |head::tail ->
	 begin 
	   if ((String.compare (List.nth head 1) "->") == 0) then
	     begin
	       let _ = output_string fp ((List.nth head 0)^" -> "^(List.nth head 2)^" ") in
	       let rec writecond = function
		 |h::t -> output_string fp (h^" "); writecond t
		 |[] -> () in
		 if ((List.length head) > 3) then
		   begin
		     let _ = output_string fp "[label = \"" in			 
		     let _ = writecond (sublist head 3 ((List.length head)-1)) in
		       output_string fp "\"]" 
		   end;
		 output_string fp "\n"
	     end  
	 end;
	  writerelations tail 
      |[] -> output_string fp "}\n"
    in
      writerelations relationscond;
      close_out fp
  in
  let _ = print_endline "generating intermediate graph" in
  let _ = print_endline ("Result of command: " ^ (string_of_int (Sys.command "dot -T pdf -v < graphscript.dot > intermediate.pdf 2> stderr.out"))) in
  let rec getarrows acc = function 
    |h::t -> 
       if (String.compare (List.nth h 1) "->") == 0 then 
	 let acc = h::acc in
	   getarrows acc t
       else getarrows acc t
    |[] -> acc
  in
  let relationsarrow = getarrows [] relationscond in
  let split act rel = 
    let actrel = List.filter (fun x -> (String.compare (List.nth x 0) act) == 0) rel in
      if (List.length actrel) < 2 then -1 (*this action has only 1 outgoing arrow*)
      else (*this action has multiple outgoing arrows thus need to add split node*)
	begin 
	  let actcond = List.flatten (List.map (fun x -> if (List.length x) < 4 then [] else sublist x 3 ((List.length x) -1)) actrel) in
	    if isemptylist actcond then 1 (*all arrows are unconditioned add AND-split*)
	    else 
	      begin
		if overlaplist actcond then 3 (*add OR-split*)
		else 2 (*add XOR-split*)
	      end
	end  
  in 
  let rec testactsplit rel = function
    |h::t ->
       let _ = match (split h rel) with
	   -1 -> print_endline (h^" doesn't need to be splitted")
	 |1 -> print_endline (h^" needs to be splitted by AND")
	 |2 -> print_endline (h^" needs to be splitted by XOR")
	 |3 -> print_endline (h^" needs to be splitted by OR") 
	 |_ -> print_endline (h^" invovles undefined split") in
	 testactsplit rel t
    |[] -> () 
  in
  let rec isopt act = function
    |h::t -> 
       if (String.compare h act) == 0 then true
       else isopt act t
    |[] -> false
  in 
  let rec preact acc = function
    |h::t -> 
       let acc = (List.nth h 0)::acc in
	 preact acc t
    |[] -> acc
  in 
  let rec ismandlist = function
    |h::t ->
       if isopt h optacts then false
       else ismandlist t
    |[] -> true
  in 
  let rec allcond = function
    |h::t -> 
       if (List.length h) < 4 then false
       else allcond t
    |[] -> true
  in 
  let join act rel = 
    let actrel = List.filter (fun x -> (String.compare (List.nth x 2) act) == 0) rel in
      if (List.length actrel) < 2 then -1 (*this action has only 1 incoming arrow*)
      else (*multiple incoming arrows, join node needed*)
	begin
	  let actcond = List.flatten (List.map (fun x -> if (List.length x) < 4 then [] else sublist x 3 ((List.length x) -1)) actrel) in
	    if isemptylist actcond then (*all incoming arrows are unconditioned*)
	      begin
		if isopt act optacts then 3 (*this action is optional, add OR-join*)
		else (*this action is mandatory*)
		  begin
		    let pact = preact [] actrel in
		      if ismandlist pact then 1
		      else 3
		  end 
	      end 
	    else  
	      begin
		if allcond actrel then (*all arrows into this action are conditioned*)
		  begin
		    let pact = preact [] actrel in
		      if ismandlist pact then 1
		      else 3
		  end 
		else 3
	      end 
	end 
  in   
  let rec testactjoin rel = function
    |h::t ->
       let _ = match (join h rel) with
	   -1 -> print_endline (h^" doesn't need to be joined")
	 |1 -> print_endline (h^" needs to be joined by AND")
	 |3 -> print_endline (h^" needs to be joined by OR")
	 |_ -> print_endline (h^" invovles undefined join") in
	 testactjoin rel t
    |[] -> ()
  in 
  let replacesplitnode nodetype act rel = 
    let rest = List.filter (fun x -> (String.compare (List.nth x 0) act) != 0) rel in
    let actrel = List.filter (fun x -> (String.compare (List.nth x 0) act) == 0) rel in
    let rec replace acc = function
      |h::t -> 
	 let acc = ((nodetype^"split"^(List.nth h 0))::(sublist h 1 ((List.length h) -1)))::acc in 
	   replace acc t
      |[] -> acc 
    in 
      (replace [] actrel)@rest
  in
  let replacejoinnode nodetype act rel = 
    let rest = List.filter (fun x -> (String.compare (List.nth x 2) act) != 0) rel in
    let actrel = List.filter (fun x -> (String.compare (List.nth x 2) act) == 0) rel in
    let rec replace acc = function
      |h::t ->
	 if (List.length h) > 3 then
	   let acc = ((sublist h 0 1)@((nodetype^"join"^(List.nth h 2))::(sublist h 3 ((List.length h) -1))))::acc in 
	     replace acc t
	 else  
	   let acc = ((sublist h 0 1)@((nodetype^"join"^(List.nth h 2))::[]))::acc in 
	     replace acc t
      |[] -> acc 
    in 
      (replace [] actrel)@rest
  in
  let rec addbpmnnodes nodes rel = function
    |h::t ->
       let addsplitnodes n act r = match (split act r) with 
 	   -1 -> n
	 |1 -> let n = ("ANDsplit"^act)::n in n
	 |2 -> let n = ("XORsplit"^act)::n in n
	 |3 -> let n = ("ORRsplit"^act)::n in n
       in 
       let nodes = addsplitnodes nodes h rel in
       let addjoinnodes n act r = match (join act r) with
  	   -1 -> n
	 |1 -> let n = ("ANDjoin"^act)::n in n
	 |3 -> let n = ("ORRjoin"^act)::n in n
       in   
       let nodes = addjoinnodes nodes h rel in  
	 addbpmnnodes nodes rel t
    |[] -> nodes 
  in
  let bpmnnodes = addbpmnnodes [] relationsarrow allactions in
  let rec bpmn acc rel = function
    |h::t ->
       let _ = print_endline ("testing "^h^" split") in
       let addsplitrel act r rel = match (split act rel) with 
 	   -1 -> let _ = print_endline (h^" doesn't need to be splitted") in r
	 |1 -> 
	    let r = ([act;"->";("ANDsplit"^act)])::(replacesplitnode "AND" act r) in 
	    let _ = print_endline (h^" has been added AND split") in r 
	 |2 ->  
	    let r = ([act;"->";("XORsplit"^act)])::(replacesplitnode "XOR" act r) in
	    let _ = print_endline (h^" has been added XOR split") in r
	 |3 -> 
	    let r = ([act;"->";("ORRsplit"^act)])::(replacesplitnode "ORR" act r) in
	    let _ = print_endline (h^" has been added OR split") in r
       in 
       let acc = addsplitrel h acc rel in
       let _ = print_endline ("testing "^h^" join") in
       let addjoinrel act r rel = match (join act rel) with
  	   -1 -> let _ = print_endline (h^" doesn't need to be joined") in r
	 |1 -> 
	    let r = ([("ANDjoin"^act);"->";act])::(replacejoinnode "AND" act r) in 
	    let _ = print_endline (h^" has been added AND join") in r
	 |3 -> 
	    let r = ([("ORRjoin"^act);"->";act])::(replacejoinnode "ORR" act r) in 
	    let _ = print_endline (h^" has been added OR join") in r
       in   
       let acc = addjoinrel h acc rel in 
	 bpmn acc rel t
    |[] -> acc 
  in
  let bpmnrel = bpmn relationsarrow relationsarrow allactions in
  let writebpmnrel =  
    let _ = print_endline "writing BPMN relations" in
    let fp = open_out_gen [Open_wronly; Open_append; Open_creat] 0o600 "output.bpm" in
      output_string fp "\nBPMN relations are:\n";
      outputlistlist fp bpmnrel;
      output_string fp "\n\n";
      close_out fp
  in
  let writegraphscript = 
    let _ = print_endline "writing graph script" in
    let fp = open_out "bpmngraphscript.dot" in
    let _ = output_string fp "digraph G{\n rankdir = LR\n" in
    let _ = output_string fp "start [label = \"start\", shape = circle]\n" in
    let _ = output_string fp "end [label = \"end\", shape = circle, style = bold]\n" in
    let rec writeactions = function
      |h::t ->
	 output_string fp (h^" [label = \""^h^"\", shape = box, style = rounded]\n");
	  writeactions t
      |[] -> () in
    let _ = writeactions actions in
    let rec writebpmnnodes = function
      |h::t ->
	 begin
	   if (String.compare (String.sub h 0 3) "AND") == 0 then
	     let _ = output_string fp (h^" [label = \"+\", shape = diamond]\n") in
	       writebpmnnodes t
	   else 
	     begin 
	       if (String.compare (String.sub h 0 3) "XOR") == 0 then
		 let _ = output_string fp (h^" [label = \"X\", shape = diamond]\n") in
		   writebpmnnodes t
	       else 
		 let _ = output_string fp (h^" [label = \"O\", shape = diamond]\n") in
		   writebpmnnodes t
	     end 
	 end
      |[] -> () in
    let _ = writebpmnnodes bpmnnodes in
    let rec writerelations = function
      |head::tail ->
	 begin 
	   if ((String.compare (List.nth head 1) "->") == 0) then
	     begin
	       let _ = output_string fp ((List.nth head 0)^" -> "^(List.nth head 2)^" ") in
	       let rec writecond = function
		 |h::t -> output_string fp (h^" "); writecond t
		 |[] -> () in
		 if ((List.length head) > 3) then
		   begin
		     let _ = output_string fp "[label = \"" in			 
		     let _ = writecond (sublist head 3 ((List.length head)-1)) in
		       output_string fp "\"]" 
		   end;
		 output_string fp "\n"
	     end  
	 end;
	  writerelations tail 
      |[] -> output_string fp "}\n"
    in
      writerelations bpmnrel;
      close_out fp
  in
  let _ = print_endline "generating BPMN graph" in
  let _ = print_endline ("Result of command: " ^ (string_of_int (Sys.command "dot -T pdf -v < bpmngraphscript.dot > bpmn.pdf 2> stderr.out"))) in
  let finishtime = Unix.gettimeofday () in
    print_endline ("Time used: "^(string_of_float (finishtime -. starttime))^" seconds")

